package org.paw.paranoidProxy;

/*
 * imports...
 */
import gnu.regexp.RE;
import gnu.regexp.REException;
import gnu.regexp.REMatch;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URISyntaxException;

import sunlabs.brazil.filter.Filter;
import sunlabs.brazil.server.Handler;
import sunlabs.brazil.server.Request;
import sunlabs.brazil.server.Server;
import sunlabs.brazil.util.http.MimeHeaders;


/**
 * this is the {@link Filter} class of the paranoidProxy. it rewrites the contents of
 * documents fetched by the proxy: all URLs found are 'tagged' with a ticket so they are allowed 
 * to pass the proxy when they are requested by a client.
 * class that rewrites URLs in fetched documents. implements the {@link Filter} interface
 * (which is a sub-interface of {@link Handler}).
 * 
 * @see Filter
 * @see ParanoidHandlerFilter
 * @see ParanoidURL
 * 
 * @author mrfox
 * @version 0.2.2
 *
 */

/*
 * HISTORY:
 * 0.2:
 * + cookie tickets added
 * + support for <BASE> tags added.
 * 
 * 0.2.1
 * * changed RE to also match google's crippled pseudo-HTML syntax
 * 
 * 0.2.2
 * - updated RE
 */

public final class ParanoidFilter extends ParanoidHandlerFilter implements Filter{
	
	/**
	 * the filter method is applied to the document's content, given the shouldFilter method called by the 
	 * filterHandler returned true.<br/>
	 * the {@link ParanoidFilter#rewriteURLs(String, String)} method is called with the document content.
	 * 
	 * @see sunlabs.brazil.filter.Filter#filter(sunlabs.brazil.server.Request, sunlabs.brazil.util.http.MimeHeaders, byte[])
	 */
	public byte[] filter(Request request, MimeHeaders headers, byte[] content)
	{	
		String uncompressHeader = headers.get("X-PAW-DECOMPRESSED");       
		
		// Do not handle if content ist still packed
		if ( uncompressHeader != null && uncompressHeader.equals("false") )
			return content;
		
		return rewriteURLs(new String(content),request.url);
	}
	
	/**
	 * this method decides whether or not the rewriting filter should be applied to a received
	 * document's content. it will be applied to all html pages.<p>
	 * if a redirect (http code 3xx or 201) is received, it is tagged with the a ticket and forwarded to the client.<p>
	 * if a cookie is present in the response, it is also tagged with a ticket.
	 * 
	 * @see sunlabs.brazil.filter.Filter#shouldFilter(sunlabs.brazil.server.Request, sunlabs.brazil.util.http.MimeHeaders)
	 */
	public boolean shouldFilter(Request request, MimeHeaders headers)
	{
		// test if we just got a REDIRECT from the web server, 
		// if so, tag it with ticket (redirect is an ABSOLUTE url)
		String redirect;
		if((redirect = headers.get("Location")) != null){
			server.log(Server.LOG_DIAGNOSTIC,prefix,"redirect detected: "+request.url+" -> "+redirect);
			ParanoidURL redirectURL = null;
			try {
				redirectURL = new ParanoidURL(redirect);
			} catch (URISyntaxException e) {
				server.log(5,prefix,"redirect URL received but could not be parsed.");
			}
			headers.put("Location",redirectURL.toExternalForm());
		}
		
		// tag cookie headers with a ticket, if present.
		// cookie example: Set-Cookie: CUSTOMER=WILE_E_COYOTE; path=/; expires=Wednesday, 09-Nov-99 23:12:40 GMT
		String cookie;		//the name/value pair of the cookie, e.g., CUSTOMER=WILE_E_COYOTE
		if((cookie = headers.get("set-cookie")) != null){
			String cookieInfo = ""; 	//path and date information, e.g., path=/; expires=Wednesday, 09-Nov-99 23:12:40 GMT
			int p;
			if((p = cookie.indexOf(';'))>0){
				cookieInfo = cookie.substring(p);
				cookie = cookie.substring(0,p);
			}
			server.log(Server.LOG_DIAGNOSTIC,prefix,"cookie detected: "+cookie+cookieInfo);
			headers.put("Set-Cookie",ParanoidURL.appendTicket(cookie)+cookieInfo);
		}
		
		String type = headers.get("content-type");	
		return (type != null && type.startsWith("text/html"));
	}
	
	/**
	 * This is the request object before the content was fetched
	 */
	public boolean respond(Request request) {
		return false;
	}
	
	/** 
	 * the most important method of this class. the document's content is parsed for URLs;
	 * all URLs found are 'tagged' with the 'ticket' that grants access to the proxy.
	 * the ticket is computed with the {@link ParanoidURL} class. the {@link gnu.regexp.RE} regular expressions package
	 * is used to find and tag links.
	 * 
	 * @param content	the content of the document to be parsed
	 * @return 			the document with all URLs found rewritten as a byte[]	
	 * 
	 * @see ParanoidURL
	 */
	byte[] rewriteURLs(String content, String baseURL){
		
		RE re = null;
		REMatch match = null;
		String docBase = null;
		
		/* write to Stream, as the string += method gets extremely slow if called for every link
		 * in a large document;
		 * initialze with 2*size of content which should be enough in most cases.
		 * (if not, it grows automatically...)
		 */
		ByteArrayOutputStream output = new ByteArrayOutputStream(2*content.length());
		
		// [scheme:][//authority][path][?query][#fragment]
		
		try {
			re = new RE("<base *href *= *\"([^\"]*)\">", RE.REG_ICASE|RE.REG_DOT_NEWLINE);
			if((match = re.getMatch(content)) != null)
				docBase = match.toString(1);
			
		} catch (REException e2) {
			server.log(Server.LOG_ERROR,prefix,"invalid <BASE> regExp in ParanoidFilter.java");
		}
		ParanoidURL base = null;
		try{
			if(docBase != null){
				try{
					base = new ParanoidURL(docBase);
				} catch (URISyntaxException e) {
					server.log(Server.LOG_WARNING,prefix,"invalid <BASE> URL found, assuming request URL as base.");
					base = new ParanoidURL(baseURL);
				}
			}else{
				base = new ParanoidURL(baseURL);
			}	
		} catch (URISyntaxException e) {
			server.log(Server.LOG_ERROR,prefix,"base URI could not be resolved, syntax error.");
			e.printStackTrace();
			// do not rewrite content.
			return content.getBytes();
		}
		
		try{
			// RE matching links in HTML documents.
			
			//re = new RE("(?: href *=| src *=|@import url\\() *\"([^\"]*)\"", RE.REG_ICASE|RE.REG_DOT_NEWLINE);
			re = new RE("( href *=| src *=|@import url\\() *\"?([^\" >]*)\"?", RE.REG_ICASE|RE.REG_DOT_NEWLINE);
			
			int index = 0;
			while((match = re.getMatch(content,index)) != null){
				
				// append text w/o matches to output
				output.write((content.substring(index,match.getStartIndex())).getBytes());
				// append first subexpression (href = ... | src = ... | ... ) to output
				output.write((match.toString(1)).getBytes());
				// append modified url to output
				try {	
					ParanoidURL cur = new ParanoidURL(match.toString(2));
					// resolve against base to get absolute URL
					cur = base.resolve(cur);
					//server.log(Server.LOG_DIAGNOSTIC,prefix,"URL "+cur.toExternalForm()+" found.");
					output.write(("\"" + cur.toExternalForm() + "\"").getBytes());
				} catch (URISyntaxException e1) {
					// found URL seems to be invalid, do nothing
				}
				// continue....
				index = match.getEndIndex();
			}
			// append rest of input
			output.write((content.substring(index,content.length())).getBytes());
		}catch (REException reEx) {
			reEx.printStackTrace();
			// stop parsing this page
			return content.getBytes();
		} catch (IOException e) {
			// happens if stream closed, should not be the case here
			server.log(Server.LOG_ERROR,prefix,"error in ParanoidFilter.java(172): output stream closed?");
			// return unprocessed content
			return content.getBytes();
		}
		
		return output.toByteArray();
	}
}