package org.paw.paranoidProxy;

import gnu.regexp.RE;
import gnu.regexp.REException;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;

import sunlabs.brazil.server.Handler;
import sunlabs.brazil.server.Request;
import sunlabs.brazil.server.Server;
import sunlabs.brazil.util.http.MimeHeaders;

/**
 * this is the {@link Handler} class of the paranoidProxy. It decides whether or not
 * incoming requests are granted.
 * Before being forwarded, all request headers are checked. Header fields are either 
 * replaced by default values, removed, or forwarded. (see documentation.)
 * 
 * @see ParanoidURL
 * @see ParanoidHandlerFilter
 * @see sunlabs.brazil.server.Handler
 * 
 * @author mrfox
 * @version 0.3.2
 *
**/
 /* HISTORY:
 * 0.2:
 * + changed URL format (see ParanoidURL)
 * + added 304 html page
 * 
 * 0.3
 * + added sanitizeHeaders() function
 * - fixed bug: request with valid ticket but containing POST data also must be blocked!
 * 
 * 0.3.1
 * - regexp for header checks updated
 * 
 * 0.3.2
 * - did not correctly terminate in case of request methods other than GET or POST
 **/
 
public final class ParanoidHandler extends ParanoidHandlerFilter implements Handler{
	
	/** parameter name in Handler configuration specifying error file name (this file is sent to the client if a request is denied)**/
	public static final String ERR403FILE = "errFile";
	
	/** name of the config parameter specifying the default accept charset (the header field will be replaced with this parameters value). **/
	public static final String DEF_ACC_CHARSET = "defaultAcceptCharset";
	/** name of the config parameter specifying the default accept encoding (the header field will be replaced with this parameters value). **/
	public static final String DEF_ACC_ENCODING = "defaultAcceptEncoding";
	/** name of the config parameter specifying the default user agent (the header field will be replaced with this parameters value). **/
	public static final String DEF_USER_AGENT = "defaultUserAgent";
	
	/** the {@link RuleManager} object containing rules to be applied to requests **/
	RuleManager rm;
	
	/** html file sent to the client in case of a HTTP 403 error **/
	String errMsg403;
	
	/** default values for header fields **/
	public static final HashMap defaultHeaders = new HashMap();
	
	public boolean init(Server server,String prefix){
		this.server = server;
		this.prefix = prefix;
		
		// query manager object
		rm = new RuleManager(server,prefix);
		
		// error file setup
		initErrMsg(server,prefix);
		
		// default headers setup
		initDefaultHeaders();
		
		// default init tasks
		return super.init(server,prefix);
	}
	
	/**
	 * decides if a request is allowed. rules are applied in following order:<p>
	 *  if a rule denying the request is found, it is refused.<p>
	 *  if a rule allowing the request is found, it is granted.<p>
	 *  if the request has a valid ticket, it is granted.<p/>
	 *  the request is refused.<p>
	 * 
	 * @return 	false, if the request is allowed. (the request is then passed to the next handler object.)
	 * 			true, if the request is not allowed. the method sends an error message to the client before returning in this case.
	 */
	public boolean respond(Request request) throws IOException {
		
		// only GET and POST methods are allowed.
		if((request.method.compareToIgnoreCase("get")!=0) && (request.method.compareToIgnoreCase("post")!=0)){
			send403(request, "only GET and POST methods are allowed");
			return true;
		}
		
		// get whole URL, without GET or POST and HTTP/1.1 or HTTP/1.0 but with query part
			//String reqStr  = request.toString();
			//String theURL = reqStr.substring(reqStr.indexOf(' ')+1,reqStr.lastIndexOf(' '));
		String theURL = request.url;
		if(request.query != "")
			theURL += "?" + request.query;
		
		// remove ticket from request, if present
		request.url = ParanoidURL.removeTicket(request.url);
		// ticket may also accur after query part
		request.query = ParanoidURL.removeTicket(request.query);
		
		server.log(Server.LOG_DIAGNOSTIC,prefix,"checking "+theURL);
		
		// does a denying rule exist?
		if(rm.denyRequest(request)){
			server.log(Server.LOG_INFORMATIONAL,prefix,theURL+" blocked (denying rule found.)");
			send403(request,"A rule denying access to the requested URL was found.");
			return true;	//stop handling that request!
		} 
		
		// allowing rule present?
		if(rm.allowRequest(request)) {
			server.log(Server.LOG_INFORMATIONAL,prefix,request.toString()+" allowed (allowing rule found.)");
			sanitizeHeaders(request);
			return false;	// request will be forwarded to next handler
		}
		
		// if POST data present, request is to be blocked if no allowing rule found
		// (even if tagged with valid ticket!)
		if(request.postData != null){
			server.log(Server.LOG_INFORMATIONAL,prefix,theURL+" blocked (unallowed POST data.)");
			send403(request,"You are not allowed to send POST data to this URL.");
			return true;	//stop handling that request!
		}
			
		// is a valid ticket present?
		// (ticket also covers GET data, if present.)
		if(ParanoidURL.checkTicket(theURL)){
			server.log(Server.LOG_INFORMATIONAL,prefix,theURL+" allowed (valid ticket found.)");
			sanitizeHeaders(request);
			return false;	// valid ticket present! request will be forwarded to next handler	
		}
		
		// default is to block request.
		send403(request,"No rule allowing your request was found. Sorry :(");
		return true;	
	}
	
	
	/**
	 * cleans request headers for requests that are forwarded. Depending on the header field type,
	 * the header field is either:<p>
	 *  removed<br>
	 *  replaced by a default value<br>
	 *  forwarded after checking its correctness
	 * 
	 * @param request	the request whose headers are sanitized.
	 */
	private void sanitizeHeaders(Request request) {
		MimeHeaders headers = request.headers;
		// walk through the headers
		
		for (int i = 0; i < headers.size(); i++) {
			String key = headers.getKey(i);
			String value = headers.get(i);
						
			// headers that are replaced by defaults.
			if(key.compareToIgnoreCase("accept-charset")==0){
				// replace by default
				String def = (String)defaultHeaders.get(new String("accept-charset"));
				if(def != "null"){
					headers.put(key,def);
					server.log(Server.LOG_DIAGNOSTIC,prefix,"accept-charset header replaced by default.");
				} else {
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"accept-charset header removed (default missing).");
				}	
			} 
			
			else if(key.compareToIgnoreCase("accept-encoding")==0){
				// replace by default
				String def = (String)defaultHeaders.get(new String("accept-encoding"));
				if(def != "null"){
					headers.put(key,def);
					server.log(Server.LOG_DIAGNOSTIC,prefix,"accept-encoding header replaced by default.");
				} else {
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"accept-encoding header removed (default missing).");
				}	
			} 
			
			else if(key.compareToIgnoreCase("user-agent")==0){
				// replace by default
				String def = (String)defaultHeaders.get(new String("user-agent"));
				if(def != "null"){
					headers.put(key,def);
					server.log(Server.LOG_DIAGNOSTIC,prefix,"user-agent header replaced by default.");
				} else {
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"user-agent header removed (default missing).");
				}
			}
			
			// headers that are forwarded after syntax check. if not correct, remove header
			// and let remote resource handle that case.
			else if(key.compareToIgnoreCase("host")==0){
				try {
					// syntax check - fqdn as defined in rfc 1034 section 3.5
					RE re = new RE("^(?:[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}|(?:[a-z0-9](?:[a-z0-9\\-])*[a-z0-9])(?:\\.[a-z0-9](?:[a-z0-9\\-])*[a-z0-9])*(?:\\.[a-z](?:[a-z0-9\\-])*[a-z0-9]))$", RE.REG_ICASE);
					if(!re.isMatch(value)){
						// invalid syntax for host name, so remove
						headers.remove(key); i--;
						server.log(Server.LOG_DIAGNOSTIC,prefix,"host header removed (invalid syntax).");
					}
				} catch (REException e) {
					e.printStackTrace();
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"host header removed (could not check syntax - invalid RE in ParanoidHandler!).");
				}
				
			} 
			
			else if(key.compareToIgnoreCase("content-length")==0){
				try {
					// syntax check - only allow digit(s)
					RE re = new RE("^[0-9][0-9]*$", RE.REG_ICASE);
					if(!re.isMatch(value)){
						// invalid syntax for host name, so remove
						headers.remove(key); i--;
						server.log(Server.LOG_DIAGNOSTIC,prefix,"content length header removed (invalid syntax).");
					}
				} catch (REException e) {
					e.printStackTrace();
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"content header removed (could not check syntax - invalid RE in ParanoidHandler!).");
				}
			}
			
			else if(key.compareToIgnoreCase("content-md5")==0){
				try {
					// syntax check - only allow valid md5 checksum
					RE re = new RE("^[0-9a-f]{32}$", RE.REG_ICASE);
					if(!re.isMatch(value)){
						// invalid syntax for host name, so remove
						headers.remove(key); i--;
						server.log(Server.LOG_DIAGNOSTIC,prefix,"content-md5 header removed (invalid syntax).");
					}
				} catch (REException e) {
					e.printStackTrace();
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"content-md5 header removed (could not check syntax - invalid RE in ParanoidHandler!).");
				}
			}
			
			else if(key.compareToIgnoreCase("pragma")==0){
				try {
					// syntax check - only allow 'no-cache' (violates rfc2616!)
					RE re = new RE("no-cache", RE.REG_ICASE);
					if(!re.isMatch(value)){
						// invalid syntax for host name, so remove
						headers.remove(key); i--;
						server.log(Server.LOG_DIAGNOSTIC,prefix,"pragma header removed (only 'no-cache' allowed).");
					}
				} catch (REException e) {
					e.printStackTrace();
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"pragma header removed (could not check syntax - invalid RE in ParanoidHandler!).");
				}
			}
			
			else if(key.compareToIgnoreCase("transfer-encoding")==0){
				try {
					// syntax check - only allow 'chunked' (violates rfc 2616!)
					RE re = new RE("chunked", RE.REG_ICASE);
					if(!re.isMatch(value)){
						// invalid syntax for host name, so remove
						headers.remove(key); i--;
						server.log(Server.LOG_DIAGNOSTIC,prefix,"transfer-encoding header removed (only 'chunked' allowed).");
					}
				} catch (REException e) {
					e.printStackTrace();
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"transfer-encoding header removed (could not check syntax - invalid RE in ParanoidHandler!).");
				}
			}
			
			else if(key.compareToIgnoreCase("expect")==0){
				try {
					// syntax check - only allow '100-continue'
					RE re = new RE("100\\-continue", RE.REG_ICASE);
					if(!re.isMatch(value)){
						// invalid syntax for host name, so remove
						headers.remove(key); i--;
						server.log(Server.LOG_DIAGNOSTIC,prefix,"transfer-encoding header removed (only 'chunked' allowed).");
					}
				} catch (REException e) {
					e.printStackTrace();
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"transfer-encoding header removed (could not check syntax - invalid RE in ParanoidHandler!).");
				}
			}
			
			else if(key.compareToIgnoreCase("cookie")==0){
				if(ParanoidURL.checkTicket(value)){
					// cookie has a valid ticket, so allow
					headers.put("Cookie",ParanoidURL.removeTicket(value));
					server.log(Server.LOG_DIAGNOSTIC,prefix,"cookie header allowed (valid ticket).");
				} else {
					headers.remove(key); i--;
					server.log(Server.LOG_DIAGNOSTIC,prefix,"cookie header removed (no valid ticket).");
				}
			}
			
			// all other header fields are simply removed
			else {
				headers.remove(key); i--;
				server.log(Server.LOG_DIAGNOSTIC,prefix,key+" header removed.");
			}
		}
	}
	
	/**
	 * sends an 'error 403' (access forbidden) message to the client. The message sent to 
	 * the client can be set in the file specified in the server's config file.
	 * 
	 * @param request	the request to be answered
	 * @param errMsg	the error message that will be shown to the user
	 */
	void send403(Request request, String errMsg){
		String message = errMsg403.replaceAll("\\{%url%\\}",request.url);
		message = message.replaceAll("\\{%errmsg%\\}",errMsg);
		request.setStatus(403);
		try {
			request.sendResponse(message.getBytes(),"text/html");
			server.log(Server.LOG_INFORMATIONAL,prefix,request.url+" blocked by proxy.");
		} catch (IOException e) {
			server.log(Server.LOG_ERROR,prefix,request.url+" blocked by proxy - ERROR: response could not be sent!");
		}
	}
	
	/**
	 * initializes the error file that is sent to a client if a request is not allowed. 	
	 * The error file specified in the server's configuration file 
	 * may contain the placeholders {%url%}, {%allowedurls%}, {%signature%}, {%errmsg%}.
	 * 
	 * @param server		the server object that initialized the handler.
	 * @param prefix		the prefix of this handler
	 */
	void initErrMsg(Server server, String prefix){
		File errFile;
		FileInputStream errFileIn = null;
		try{
			errFile = new File(server.props.getProperty(prefix + ERR403FILE));
			errFileIn = new FileInputStream(errFile);
		} catch (FileNotFoundException e) {
			errFile = null;
			server.log(Server.LOG_WARNING,prefix,"html error file not found.");
			// default string will be used.
		} catch (Exception e){
			errFile = null;
			server.log(Server.LOG_WARNING,prefix,"html error file not specified (parameter "+ERR403FILE+").");
			// default string will be used.
		}
		
		if(errFile != null){
			// 65k bytes should be enough for an error file
			byte[] content = new byte[(int)errFile.length()];
			try {
				errFileIn.read(content);
			} catch (Exception e) {
				errFile = null;
			}
			String message = new String(content);
			// replace placeholders
			message = message.replaceAll("\\{%allowedurls%\\}",rm.rulesAllow.printURLs());
			message = message.replaceAll("\\{%signature%\\}",SIGNATURE);
			errMsg403 = message;
		} 
		if(errFile == null)
			errMsg403 = "access to {%url%} has been blocked by the proxy server.";
		
	}
	
	/**
	 * reads the values for those headers which are replaced by a default value.
	 * the values should be specified in the {@link Handler}'s configuration. 	 *
	 */	
	void initDefaultHeaders(){
			
		String defAccChar = server.props.getProperty(prefix + DEF_ACC_CHARSET);
		String defAccEnc = server.props.getProperty(prefix + DEF_ACC_ENCODING);
		String defUA = server.props.getProperty(prefix + DEF_USER_AGENT);
		
		if(defAccChar == null){
			defAccChar = "ISO-8859-1,utf-8;q=0.7,*;q=0.7";
			server.log(Server.LOG_WARNING,prefix,"you should define a default Accept Charset in the ParanoidHandler config (<param name=\""+prefix+DEF_ACC_CHARSET+"\" value=\"...\" />)");
		}
		if(defAccEnc == null){
			defAccEnc = "gzip,deflate";
			server.log(Server.LOG_WARNING,prefix,"you should define a default Accept Encoding in theParanoidHandler config (<param name=\""+prefix+DEF_ACC_ENCODING+"\" value=\"...\" />)");
		}
		if(defUA == null){
			defUA = "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.7.5) Gecko/20050210 Firefox/1.0";
			server.log(Server.LOG_WARNING,this.prefix,"you should define a default User Agent in the ParanoidHandler config (<param name=\""+prefix+DEF_USER_AGENT+"\" value=\"...\" />)");
		}
		
		defaultHeaders.put(new String("accept-charset"),defAccChar);
		defaultHeaders.put(new String("accept-encoding"),defAccEnc);
		defaultHeaders.put(new String("user-agent"),defUA);
	}
}