package org.paw.paranoidProxy;

import java.io.File;
import java.io.IOException;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;

import sunlabs.brazil.server.Handler;
import sunlabs.brazil.server.Request;
import sunlabs.brazil.server.Server;

/**
 * abstract class implementing the {@link Handler} interface. this class is used to provide
 * code used for both the {@link ParanoidHandler} and {@link ParanoidFilter} classes of ParanoidProxy.<p/>
 * 
 * in particular it initializes the proxy secret and hash algorithm.
 * 
 * @see ParanoidFilter
 * @see ParanoidHandler
 * @see ParanoidURL
 * @see sunlabs.brazil.server.Handler
 * 
 * @author mrfox
 * @version 0.2
 */
/*
 * HISTORY
 * 0.2:
 * + readConfigParam() method added so that classes implementing this abstract class have
 *   easy access to the server config file.
 */
public abstract class ParanoidHandlerFilter implements Handler{
	
	/** the @link Server} object that initialized this {@link ParanoidHandlerFilter} object **/
	Server server;	
	/** used to get access to {@link Server#props} properties associated with this prefix **/
	String prefix;
	
	/** secret used for HMAC computation, read from server's config file **/
	String secret;
	/** algorithm used for HMAC computation, read from server's config file **/
	String algorithm;
	
	/** signature on error pages **/
	public static final String SIGNATURE = "ParanoidProxy version 0.3.1a";
	
	/**
	 * name of the config parameter specifying the location of the server configuration file. This config 
	 * parameter must be present in the configuration of a class extending this abstract class.
	 */
	public static final String CONFIG = "serverConfig";
	
	/**
	 * the init method is called when the {@link sunlabs.brazil.server.Handler} or {@link sunlabs.brazil.filter.Filter} object is created by the 
	 * http server object (on startup)<p/>
	 * the method is used to initialize the {@link sunlabs.brazil.server.Handler} or {@link sunlabs.brazil.filter.Filter}. It also initializes the static
	 * message digest object of the {@link ParanoidURL} class.
	 * 
	 * @see sunlabs.brazil.server.Handler#init(sunlabs.brazil.server.Server, java.lang.String)
	 * @see ParanoidURL
	 */	
	public boolean init(Server server, String prefix) {
		
		this.server = server;
		this.prefix = prefix;
		
		// check whether static message digest and secret of ParanoidURL class have already been initialized
		if(!ParanoidURL.isInit()){
			
			server.log(Server.LOG_DIAGNOSTIC,prefix,"initializing Message Digest.");
			
			// read server config file containing secret and algorithm to use
			String configFile = server.props.getProperty(prefix + CONFIG);
			
			secret = readConfigParam(configFile,"secret");
			if(secret == null)
				server.log(Server.LOG_ERROR,prefix,"a proxy secret must be defined in the server config file (<secret>...</secret>");
			
			algorithm = readConfigParam(configFile,"algorithm");
			if(algorithm==null){
				server.log(Server.LOG_INFORMATIONAL,prefix,"no algorithm specified in server config, using SHA1.");
				algorithm = "SHA";
			}
			
			// initialize ParanoidURL static fields
			if(!ParanoidURL.init(secret,algorithm)){
				server.log(Server.LOG_ERROR,prefix,"error initializing Message Digest.");
				return false;
			}
			
		}
		server.log(Server.LOG_DIAGNOSTIC,prefix,"--- CREATED ---");
		return true;
	}
	
	public abstract boolean respond(Request request) throws IOException;
	
	
	/**
	 * read a parameter in the server XML config file. The init method must have been called BEFORE 
	 * any calls to this method.
	 * 
	 * @param configFile	name of the XML file containing the parameter &lt;paramName&gt;
	 * @param paramName		name of the parameter to read
	 * 
	 * @return 				value of the parameter, or null if not present in config file
	 */
	String readConfigParam(String configFile, String paramName){
		
		SAXBuilder builder = new SAXBuilder(false);
		Document doc = null;
		
		try {
			File f = new File(configFile);
			if(!f.canRead()){
				server.log(Server.LOG_ERROR,prefix,"config file "+configFile+" not readable.");
				return null;
			}
			doc = builder.build(f);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		// Get root element
		Element root = doc.getRootElement();
		// return parameter value
		return root.getChildText(paramName);
	}
}