package org.paw.paranoidProxy;

import java.net.URI;
import java.net.URISyntaxException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import javax.crypto.Mac;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

/**
 * a wrapper class for java's URI class. it provides the necessary methods for re-writing URLs 
 * with a ticket (HMAC), i.e. 
 * <p>
 * &lt;scheme&gt;:&lt;authority&gt;&lt;path&gt;?&lt;query&gt;&#35;&lt;reference&gt;<br/>
 * becomes<br/>
 * &lt;scheme&gt;:&lt;authority&gt;&lt;path&gt;?&lt;query&gt;&#123;proxy_ticket&#125;&#35;&lt;reference&gt;<br/>
 * where &#123; and &#125; stand for the BEGIN_TICKET_SEPARATOR and END_TICKET_SEPARATOR constants, respectively.
 * <p>
 * (the reference part of the URL corresponds to the URIs fragment part)
 * 
 * @author 		mrfox
 * @version		0.4
 */

/*
 * history:
 * 0.2:
 * + changed URL scheme from <scheme>://<host><path>?<query>#<fragment>_ticket
 *   to
 *   <scheme>://<host><path>?<query>{ticket}#<fragment>
 *   since the browser never sends the fragment along
 * + md now static (better performance ?)
 * + BEGIN_TICKET_SEPARATOR and END_TICKET_SEPARATOR defined
 * + handling of URLs with missing '/' at the end (see toExternalForm method)
 * 
 * 0.3:
 * + added static checkTicket(), appendTicket(), and removeTicket() methods
 * - lots of bugfixes
 * 
 * 0.3.1
 * - &amp; in HTML URLs was not correctly handled.
 * 
 * 0.4
 * * replaced java.security.MessageDigest by javax.crypto.Mac which implements rfc2104
 * - added isInit() method since message digest object is private and should not be accessed from
 *   outside to check whether init() has been called.
 */

final class ParanoidURL{
	
	/* RFC 1738: *******************************************************
	 only alphanumerics, the special characters "$-_.+!*'(),", and
	 reserved characters used for their reserved purposes may be used
	 unencoded within a URL.
	 *******************************************************************/
	
	/** separator used to indicate begin of ticket within URL **/
	public static final String BEGIN_TICKET_SEPARATOR = "%7B";
	/** separator used to indicate end of ticket within URL **/
	public static final String END_TICKET_SEPARATOR = "%7D";
	/** the default algorithm to use for HMAC computation **/
	public static final String DEFAULT_HMAC_ALGO = "HmacSHA1";
	
	URI uri;
	/** the secret key used for HMAC computation. it must be initialized before use of the class. **/
	static SecretKey secret = null;
	/** the message digest object used for HMAC computation. it must be initialized before use of the class. **/
	static Mac hmac = null;
	
	
	/**
	 * initializes the {@link ParanoidURL} class by setting the secret and algorithm for the static message digest.
	 * <p>the default algorithm for the message digest is SHA and is used if the passed algorithm is unknown.
	 * 
	 * @param secret		secret for HMAC computation
	 * @param algorithm		algorithm for HMAC computation
	 * @return				true iff secret and digest are properly initialized
	 */
	public static boolean init(String secret, String algorithm){
		
		if(secret==null || secret=="")
			return false;
				
		try {
			hmac = Mac.getInstance(algorithm);
		} catch (NoSuchAlgorithmException e) {
			// set to default algo
			System.out.println("paranoidURL: "+algorithm+": unknown algorithm, using "+DEFAULT_HMAC_ALGO+".");
			algorithm = DEFAULT_HMAC_ALGO;
			try {
				hmac = Mac.getInstance(algorithm);
			} catch (NoSuchAlgorithmException e1) {
				System.out.println("ParanoidURL: unable to instantiate message digest.");
				hmac = null;
				return false;
			}
		}
		
		ParanoidURL.secret = new SecretKeySpec(secret.getBytes(),algorithm);
		
		try {
			// algorithm should now be set to a valid option
			hmac.init(ParanoidURL.secret);
		} catch (InvalidKeyException e1) {
			e1.printStackTrace();
			System.out.println("ParanoidURL: unable to initialize message digest (invalid key).");
			hmac = null;
			return false;
		}
		
		return true;
	}
	
	/**
	 * Appends the HMAC ticket computed with the secret to the given input. the ticket is enclosed
	 * by BEGIN_TICKET_SEPARATOR and END_TICKET_SEPARATOR.
	 * 
	 * @param input		the input for which the ticket is computed
	 * @return			input+BEGIN_TICKET_SEPARATOR+ticket+END_TICKET_SEPARATOR
	 */
	public static String appendTicket(String input){
		// needed to convert byte[] to hexadecimal string
		final char[] HEX_DIGITS = {'0', '1', '2', '3', '4', '5', '6',
				'7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};		
		byte[] ticket = null;
				
		// thread-safe...
		synchronized(hmac){
			ticket = hmac.doFinal(input.getBytes());
		}
		
		// convert byte[] to hex string
		StringBuffer sb = new StringBuffer();		
		for (int i = 0; i < ticket.length; i++) {
			sb.append(HEX_DIGITS[(ticket[i] & 0xF0) >> 4]);
			sb.append(HEX_DIGITS[ticket[i] & 0x0F]);
		}
		
		// now tag input with ticket
		return input + BEGIN_TICKET_SEPARATOR + sb.toString() + END_TICKET_SEPARATOR;
	}
	
	/**
	 * removes a ticket enclosed by BEGIN_TICKET_SEPARATOR and END_TICKET_SEPARATOR from the input, if present.<p>
	 * examples ('{' and '}' are begin- and end-separators, respectively):<br>
	 * &nbsp;removeTicket("foo{bar}") returns "foo".<br>
	 * &nbsp;removeTicket("foo{bar}foo") returns "foofoo".<br>
	 * &nbsp;removeTicket("foo}bar{") returns "foo}bar{".  
	 * 
	 * @param input		string possibly containing a ticket
	 * @return			input string without the ticket and the separators. 
	 */
	public static String removeTicket(String input){
		int posB = input.lastIndexOf(BEGIN_TICKET_SEPARATOR);
		int posE = input.lastIndexOf(END_TICKET_SEPARATOR);
		
		if(posB>0 && posE > posB){
			return input.substring(0,posB)+input.substring(posE+END_TICKET_SEPARATOR.length());
		} else {
			return input;
		}
	}
	
	/**
	 * checks whether the input has appended a valid ticket enclosed by BEGIN_TICKET_SEPARATOR and END_TICKET_SEPARATOR.
	 * if the ticket is not at the end of the input, false is always returned.<p>
	 * examples ('{' and '}' are begin- and end-separators, respectively):<br>
	 * &nbsp; checkTicket("foo{&lt;valid_ticket_for_foo&gt;}") returns true.<br>
	 * &nbsp; checkTicket("foo{&lt;invalid_ticket_for_foo&gt;}") returns false.<br>
	 * &nbsp; checkTicket("foo{&lt;valid_ticket_for_foo&gt;}bar") returns false.  
	 * 
	 * @param input		string possibly containing a ticket.
	 * @return			true iff the ticket is a valid HMAC for the preceding part of the input.
	 */
	public static boolean checkTicket(String input){
		String noTicketInput = removeTicket(input);
		if(noTicketInput.equals(input))
			return false;
		String computedTicket = appendTicket(noTicketInput);
		return input.equals(computedTicket);
	}
	
	
	/**
	 * constructs a {@link ParanoidURL} from a String representing an URL.<p>
	 * occurences of '&amp;' are rewritten as '&' first, since the '&' symbol is a reserved URL
	 * character, but must be encoded when used in HTML documents. 
	 * 
	 * @param url			URL to construct ParanoidURL from
	 * @throws URISyntaxException
	 */
	public ParanoidURL(String url) throws URISyntaxException{
		// the ampersand '&' character is a reserved character in URLs and must be rewritten in 
		// HTML as '&amp;'. the browser will then replace the '&amp' with '&' when requesting the URL.
		// (the same is true for '<' amd '>', but they should not occur in URLs)
		
		this.uri = new URI(url.replaceAll("&amp;","&"));
	}
	
	/**
	 * constructs a {@link ParanoidURL} from an {@link URI}
	 * 
	 * @param uri			URI to construct ParanoidURL from
	 */
	public ParanoidURL(URI uri){
		this.uri = uri;
	}
	
	/**
	 * returns a string representation of this URL object, including a ticket. the ticket is enclosed 
	 * with the BEGIN_TICKET_SEPARATOR and END_TICKET_SEPARATOR separators. 
	 * the ticket is appended after the path and query part of the URL, but before the fragment part.
	 * 
	 * @return String representation of URL w. ticket
	 */
	public String toExternalForm(){
		String input = uri.toString();
		
		/* 
		 * if no path part present in URL, append trailing '/' as most browsers will
		 * e.g., http://abc.de becomes http://abc.de/
		 * if the trailing / is missing for directories other then root, most servers will
		 * send a 'permanently moved' relocation which is handled by the filter class
		 */
		
		// for some reasons, sometimes null is returned and sometimes "" if the path part is empty...
		if(uri.getPath()==null || (uri.getPath().compareTo("")==0))
			input += "/";
		
		/*
		 * the fragment part of an URI is never sent to the proxy.
		 */
		if(uri.getFragment() != null){
			int pos = input.indexOf('#');
			return appendTicket(input.substring(0,pos))+input.substring(pos);
		}
		else return appendTicket(input);	
	}
	
	
	/**
	 * wrapper method for the {@link URI#resolve(java.net.URI)} method.
	 * 
	 * @param url	the url to be resolved, e.g. /path/to/docu.ment
	 * @return		the absolute URL, e.g. http://base.url/path/to/docu.ment
	 */
	public ParanoidURL resolve(ParanoidURL url){
		return new ParanoidURL(this.uri.resolve(url.uri));
	}
	
	
	/**
	 * check whether the {@link #init(String,String)} method has been executed.
	 * 
	 * @return		true, iff itit() method was called and message digest is initialized.
	 */
	public static boolean isInit(){
		return hmac != null;
	}
}