package org.paw.paranoidProxy;

import gnu.regexp.RE;
import gnu.regexp.REException;

import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import sunlabs.brazil.server.Request;
import sunlabs.brazil.server.Server;
import sunlabs.brazil.util.http.HttpUtil;

/**
 * this class manages a set of rules specified for the proxy. it maintains two data structures, 
 * one for rules with target "allow" and one for rules with target "deny". rules are read from an 
 * XML rule file.<p>
 * for each rule, a {@link RuleManager.Rule} object is instantiated. The {@link RuleManager.Rule}
 * object contains the Data Handler matching a given request against the rule's parameter set.
 * the rule object is then inserted into the respective data structure (allow or deny) with 
 * the respective URL(s) as the key.
 */

final class RuleManager{
	
	/**
	 * deny rule list. A request will be denied if it has a matching rule in rulesDeny, regardless of any matching rules in rulesAllow.
	 * Otherwise it is allowed iff it has a matching rule in rulesAllow.
	 */
	RuleList rulesAllow;
	/**
	 * allow rule list. A request will be denied if it has a matching rule in rulesDeny, regardless of any matching rules in rulesAllow.
	 * Otherwise it is allowed iff it has a matching rule in rulesAllow.
	 */
	RuleList rulesDeny;
	
	/**
	 * name of the config parameter specifying the location of the XML rule file. This config 
	 * parameter must be present in the {@link ParanoidHandler}'s configuration.
	 */
	public static final String RULES = "rules";	
	
	/** used to get access to {@link Server#props} properties associated with this prefix **/
	String prefix;
	/** the server object **/
	Server server;
	
	
	/**
	 * @param server	server object the RuleManager is associated with
	 * @param prefix	prefix of the {@link sunlabs.brazil.server.Handler} instantiating the RuleManager
	 */
	public RuleManager(Server server,String prefix){

		this.prefix = prefix;
		this.server = server;
		
		rulesAllow = new RuleList();
		rulesDeny = new RuleList();
		
		String ruleFile = server.props.getProperty(prefix + RULES);
		
		loadRules(ruleFile);
	}
	
	/**
	 * decide whether the given {@link Request} is to be allowed. If a matching denying rule exists, false is returned.
	 * otherwise true is returned iff the request has a matching allowing rule.
	 * 
	 * @param request	request to be checked
	 * @return			true iff request is allowed with loaded rule set.
	 */
	public boolean allowRequest(Request request){
		return(rulesAllow.isMatch(request) && !rulesDeny.isMatch(request));
	}
	
	/**
	 * decide whether given {@link Request} is to be denied. If a matching denying rule exists, true is returned.
	 * 
	 * @param request	request to be checked
	 * @return			true iff request is not allowed with loaded rule set.
	 */
	public boolean denyRequest(Request request){
		return(rulesDeny.isMatch(request));
	}
	
	/**
	 * reads all rules from the XML file specified in the XML description of the {@link sunlabs.brazil.server.Handler}
	 * that owns this RuleManager object.
	 * 
	 * @param ruleFile	name of the XML config file
	 */
	void loadRules(String ruleFile){
		
		try {
			
			SAXBuilder builder = new SAXBuilder(false);
			Document doc = null;
			
			try {
				File f = new File(ruleFile);
				if(!f.canRead()){
					server.log(Server.LOG_ERROR,prefix,"config file "+ruleFile+" not readable.");
					return;
				}
				doc = builder.build(f);
			} catch (IOException e) {
				e.printStackTrace();
			}
			
			// Get root element
			Element root = doc.getRootElement();
			
			// now insert every rule
			List rules = root.getChildren();
			Iterator it = rules.iterator();
			
			while(it.hasNext()){
				Element curRule = (Element)it.next();
				
				// if rule target is 'allow', add to rulesAllow. if 'deny', add to rulesDeny.
				RuleList target;
				if(curRule.getAttributeValue("target").equalsIgnoreCase("allow"))
					target = rulesAllow;
				else
					target = rulesDeny;
				
				// generate rule
				Rule r = new Rule(curRule.getChildText("name"),curRule.getChildText("description"));
				
				// add all parameters to rule	
				List paramList = curRule.getChildren("param");
				Iterator paramIt = paramList.iterator();
				while(paramIt.hasNext()){
					Element p = (Element)(paramIt.next());
					if(p.getAttribute("required")!=null && p.getAttributeValue("required").equalsIgnoreCase("required"))
						r.addParam(p.getChildText("pname"),p.getChildText("pvalue"),p.getAttributeValue("method"),true);
					else
						r.addParam(p.getChildText("pname"),p.getChildText("pvalue"),p.getAttributeValue("method"));
				}
				
				// add the rule for all URLs specified
				List URLs = curRule.getChildren("url");
				Iterator URLit = URLs.iterator();
				while(URLit.hasNext()){
					String url = ((Element)URLit.next()).getText();
					target.insertRule(url,r);
				}
				// at least one URL should be there...
				if(URLs.isEmpty()){
					server.log(Server.LOG_WARNING,prefix,"rule '"+r.ruleName+"': no URLs specified, rule will be ignored.");
				} else {
					server.log(Server.LOG_INFORMATIONAL,prefix,"rule '"+r.ruleName+"' added ("+r.ruleDescription+", target: "+curRule.getAttributeValue("target")+")");
				}
			}
			
		} catch (JDOMException jdome) {
			server.log(Server.LOG_ERROR,prefix,"config file "+ruleFile+": syntax error.");
			jdome.printStackTrace();
		}
		
	}
	
	/**
	 * represents a rule for an URL, consisting of parameters and legal values for them.
	 * a request containing client-provided data may only be allowed if the {@link #isMatch(Request)} 
	 * method returns true.<p>
	 * the Rule object consists of a Name, a Description, and the Client Data Handler along with a
	 * parameter set as its input.
	 */
	class Rule{
		/** a name for informational purposes **/
		public String ruleName;
		/** a description for informational purposes **/
		public String ruleDescription;
		
		/** GET parameter names of the Client Data Handler **/
		Vector GETparams;
		/** POST parameter names of the Client Data Handler **/
		Vector POSTparams;
		/** allowed values for GET parameters of the Client Data Handler (regular expression)**/
		Vector GETvalues;
		/** allowed values for POST parameters of the Client Data Handler (regular expression)**/
		Vector POSTvalues;
		/** required GET parameters of the Client Data Handler**/
		Vector GETrequired;	
		/** required POST parameters of the Client Data Handler**/
		Vector POSTrequired;
		
		/**
		 * creates a new Rule object with an empty parameter set (i.e., the data handler will block 
		 * all requests containing client provided data).
		 * 
		 * @param name			rule name, may be null
		 * @param description	rule description, may be null
		 */
		public Rule(String name,String description){
			if(name==null)
				ruleName = "[name missing]";
			else
				ruleName = name;
			
			if(description==null)
				ruleDescription = "";
			else
				ruleDescription = description;
			
			GETparams = new Vector();
			GETvalues = new Vector();
			GETrequired = new Vector();
			POSTparams = new Vector();
			POSTvalues = new Vector();
			POSTrequired = new Vector();
		}
		
		/**
		 * creates a new Rule object with an empty parameter set and empty description.
		 * 
		 * @param name			rule name, may be null
		 */
		public Rule(String name){
			new Rule(name,null);
		}
		
		/**
		 * add a parameter to this rule's parameter set.
		 * 
		 * @param name			name of the parameter
		 * @param value			legal values (regular expression)
		 * @param method		http method (GET or POST)
		 * @param isRequired	true iff the parameter may not be omitted
		 */
		public void addParam(String name,String value,String method, boolean isRequired){
			
			if(name==null || value==null || method== null){
				server.log(Server.LOG_WARNING,prefix,"rule "+ruleName+": parameter name, value, and method must not be empty. parameter ignored.");	
				return;
			}
			
			// generate regular expression representing allowed values
			RE cur = null;
			try{
				cur = new RE(value);
			} catch (REException e) {
				// expression in rule file is not valid, just ignore this rule and continue, log error
				server.log(Server.LOG_WARNING,prefix,"rule "+ruleName+": "+value+" is not a valid regular expression - parameter ignored.");
				return;
			}
			
			// add to GET or POST vectors
			if(method.compareToIgnoreCase("get")==0){
				GETparams.add(name);
				GETvalues.add(cur);
				if(isRequired)
					GETrequired.add(name);
				
			} else if (method.compareToIgnoreCase("post")==0){
				POSTparams.add(name);
				POSTvalues.add(cur);
				if(isRequired)
					POSTrequired.add(name);
				
			} else
				server.log(Server.LOG_WARNING,prefix,"rule "+ruleName+": invalid HTTP method (only GET and POST allowed) - parameter ignored.");
		}
		
		/**
		 * wrapper method for {@link #addParam(String, String, String, boolean)}. isRequired defaults to false.
		 * 
		 * @param name			name of the parameter
		 * @param value			legal values (regular expression)
		 */
		public void addParam(String name,String value,String method){
			addParam(name,value,method,false);
		}
		
		/**
		 * this is the Data Handler matching a given request vs. a rule's parameter set.
		 * it checks whether a given request complies with rules specified for client-provided data.
		 * the following checks are done:<p>
		 *  all required parameters are there.<br>
		 *  no unallowed parameters are there.<br>
		 *  parameter values are in a legal range.
		 *  
		 * @param r				the request to be checked.
		 * @return	boolean		true, iff request is to be allowed
		 */
		public boolean isMatch(Request r){
			
			Iterator it;
			
			// read data into arrays.
			Hashtable POSTdata = new Hashtable();
			if(r.postData != null){
				HttpUtil.extractQuery(new String(r.postData),POSTdata);
			}
			
			Hashtable GETdata = new Hashtable();
			if(r.query != null){
				HttpUtil.extractQuery(new String(r.query),GETdata);
			}
			
			// DATA HANDLER ////////////////////////////////////
			
			/*
			 * first, check if all required parameters are here.
			 */
			it = GETrequired.iterator();
			while(it.hasNext()){
				String name = (String)it.next();
				if((GETdata.get(name)) == null)
					return false;
			}
			
			it = POSTrequired.iterator();
			while(it.hasNext()){
				String name = (String)it.next();
				if((POSTdata.get(name)) == null)
					return false;
			}
						
			/*
			 * then check all parameter values for correctness. remove processed parameters.
			 */	
			
			server.log(Server.LOG_DIAGNOSTIC,prefix,"checking "+GETdata.size()+" GET parameters...");

			for(int i=0; i<GETparams.size(); i++){
				// get parameter name
				String name = (String)GETparams.get(i);
				// get client-provided parameter value, if present
				String val;
				if((val = (String)GETdata.get(name)) != null){
					// value has been sent, check if correct:
					
					if( ((RE)GETvalues.get(i)).getMatch(val) == null ){
						// passed value DOES NOT MATCH!!!
						server.log(Server.LOG_DIAGNOSTIC,prefix,"checking parameter "+name+": match "+val+" against "+((RE)GETvalues.get(i)).toString()+"... FAILED.");
						return false;
					} else {
						// value OK, remove from list and continue.
						server.log(Server.LOG_DIAGNOSTIC,prefix,"checking parameter "+name+": match "+val+" against "+((RE)GETvalues.get(i)).toString()+"... ok.");
						GETdata.remove(name);
					}
				} else
					server.log(Server.LOG_DIAGNOSTIC,prefix,r.url+": parameter "+name+" not present (and not required,thus ok.)");
			}
			
			// if hashtable with values not empty, client tried to push additional parameters, block request.
			if(GETdata.size() >0)
				return false;
			
			
			server.log(Server.LOG_DIAGNOSTIC,prefix,"checking "+POSTdata.size()+" POST parameters...");

			for(int i=0; i<POSTparams.size(); i++){
				// get parameter name
				String name = (String)POSTparams.get(i);
				// get client-provided parameter value, if present
				String val;
				if((val = (String)POSTdata.get(name)) != null){
					// value has been sent, check if correct:
					
					if( ((RE)POSTvalues.get(i)).getMatch(val) == null ){
						// passed value DOES NOT MATCH!!!
						server.log(Server.LOG_DIAGNOSTIC,prefix,"checking parameter "+name+": match "+val+" against "+((RE)POSTvalues.get(i)).toString()+"... FAILED.");
						return false;
					} else {
						// value OK, remove from list and continue.
						server.log(Server.LOG_DIAGNOSTIC,prefix,"checking parameter "+name+": match "+val+" against "+((RE)POSTvalues.get(i)).toString()+"... ok.");
						POSTdata.remove(name);
					}
				} else
					server.log(Server.LOG_DIAGNOSTIC,prefix,r.url+": parameter "+name+" not present (and not required,thus ok.)");
			}
			
			// if hashtable with values not empty, client tried to push additional parameters, block request.
			if(POSTdata.size() >0)
				return false;		
			
			return true;
		}
	}
	
	/**
	 * manages rules per URLs. several rules may be present for a single URL.
	 */
	class RuleList{
		Hashtable rules;
		
		public RuleList(){
			rules = new Hashtable();
		}
		
		/**
		 * inserts a rule for an URL. several rules may be specified for one url. 
		 * 
		 * @param rule	the rule object representing the rule
		 * @param url	the URL the rule is valid for
		 */
		public void insertRule(String url,Rule rule){
			// check if other rules for this URL exist
			if(rules.containsKey(url)){
				// url present, append rule
				((Vector)rules.get(url)).add(rule);
			} else {
				// no rules for this URL yet
				Vector v = new Vector();
				v.add(rule);
				rules.put(url,v);
			}
		}
		
		/**
		 * checks whether a request is allowed for a given URL. several rules for client-provided data may exist for one URL, and one rule may apply to several URLs.
		 * 
		 * @param request		the request to be checked
		 * @return				true iff a rule exists that returned true
		 */
		public boolean isMatch(Request request){
			
			server.log(Server.LOG_DIAGNOSTIC,prefix,"checking "+request.url+"...");
			if(!rules.containsKey(request.url)){
				server.log(Server.LOG_DIAGNOSTIC,prefix,"no rule for "+request.url +" found.");
				return false;
			} else {
				Iterator it = ((Vector)rules.get(request.url)).iterator();
				while(it.hasNext()){
					if( ((Rule)it.next()).isMatch(request) )
						return true;
				}
				return false;
			}
		}
		
		/**
		 * prints a list of all allowed URLs to be sent to a client.
		 * 
		 * @return	List of allowed URLs.
		 */
		public String printURLs(){
			Enumeration keyEn = rules.keys();
			String res = "";
			while(keyEn.hasMoreElements()){
				String url = (String)(keyEn.nextElement());
				res += "<br/><a href=\""+url+"\">"+url+"</a><br/>";
			}
			return res;
		}
	}
}
